@@ -29,10 +29,6 @@
 #include <ctype.h>
 #include <locale.h>
 
-#ifdef HAVE_ALLOCA_H
-#include <alloca.h>
-#endif
-
 #include "gmime-message.h"
 #include "gmime-utils.h"
 #include "gmime-stream-mem.h"
@@ -290,8 +286,10 @@ g_mime_message_add_recipients_from_string (GMimeMessage *message, char *type, co
 	g_hash_table_remove (message->header->recipients, type);
 	
 	addrlist = internet_address_parse_string (string);
-	if (addrlist)
+	if (addrlist) {
 		recipients = internet_address_list_concat (recipients, addrlist);
+		internet_address_list_destroy (addrlist);
+	}
 	
 	g_hash_table_insert (message->header->recipients, type, recipients);
 	
@@ -560,7 +558,7 @@ g_mime_message_set_mime_part (GMimeMessage *message, GMimePart *mime_part)
  * @message: MIME Message
  * @stream: output stream
  *
- * Write the contents of the MIME Message to @string.
+ * Write the contents of the MIME Message to @stream.
  **/
 void
 g_mime_message_write_to_stream (GMimeMessage *message, GMimeStream *stream)
@@ -714,7 +712,7 @@ g_mime_message_get_body (const GMimeMessage *message, gboolean want_plain, gbool
 	const char *content;
 	char *body = NULL;
 	size_t len = 0;
-
+	
 	g_return_val_if_fail (message!=NULL, NULL);
 	g_return_val_if_fail (is_html!=NULL, NULL);
 	
