@@ -29,6 +29,7 @@
 #include <ctype.h>
 
 #include "gmime-parser.h"
+#include "gmime-stream-mem.h"
 
 #ifndef HAVE_ISBLANK
 #define isblank(c) ((c) == ' ' || (c) == '\t')
@@ -82,8 +83,9 @@ struct _GMimeParser {
 	off_t headers_start;
 	off_t header_start;
 	
-	unsigned int unstep:31;
+	unsigned int unstep:30;
 	unsigned int midline:1;
+	unsigned int seekable:1;
 	
 	GMimeContentType *content_type;
 	struct _header_raw *headers;
@@ -228,6 +230,7 @@ parser_init (GMimeParser *parser, GMimeStream *stream)
 	
 	parser->unstep = 0;
 	parser->midline = FALSE;
+	parser->seekable = offset != -1;
 	
 	parser->headers = NULL;
 	
@@ -303,50 +306,58 @@ parser_offset (GMimeParser *parser, unsigned char *cur)
 	return (parser->offset - (parser->inend - inptr));
 }
 
-#define header_backup(parser, start, len) G_STMT_START {                    \
-	if (parser->headerleft <= len) {                                    \
-		unsigned int hlen, hoff;                                  \
-		                                                          \
+#define header_backup(parser, start, len) G_STMT_START {                      \
+	if (parser->headerleft <= len) {                                      \
+		unsigned int hlen, hoff;                                      \
+		                                                              \
 		hlen = hoff = parser->headerptr - parser->headerbuf;          \
-		hlen = hlen ? hlen : 1;                                   \
-		                                                          \
-		while (hlen < hoff + len)                                 \
-			hlen <<= 1;                                       \
-		                                                          \
+		hlen = hlen ? hlen : 1;                                       \
+		                                                              \
+		while (hlen < hoff + len)                                     \
+			hlen <<= 1;                                           \
+		                                                              \
 		parser->headerbuf = g_realloc (parser->headerbuf, hlen + 1);  \
 		parser->headerptr = parser->headerbuf + hoff;                 \
-		parser->headerleft = hlen - hoff;                           \
-	}                                                                 \
-	                                                                  \
-	memcpy (parser->headerptr, start, len);                             \
-	parser->headerptr += len;                                           \
-	parser->headerleft -= len;                                          \
+		parser->headerleft = hlen - hoff;                             \
+	}                                                                     \
+	                                                                      \
+	memcpy (parser->headerptr, start, len);                               \
+	parser->headerptr += len;                                             \
+	parser->headerleft -= len;                                            \
 } G_STMT_END
 
 #define header_parse(parser, hend) G_STMT_START {                           \
-	register unsigned char *colon;                                    \
-	struct _header_raw *header;                                       \
-	unsigned int hlen;                                                \
-	                                                                  \
-	header = g_new (struct _header_raw, 1);                           \
-	header->next = NULL;                                              \
-	                                                                  \
+	register unsigned char *colon;                                      \
+	struct _header_raw *header;                                         \
+	unsigned int hlen;                                                  \
+	                                                                    \
+	header = g_new (struct _header_raw, 1);                             \
+	header->next = NULL;                                                \
+	                                                                    \
 	*parser->headerptr = '\0';                                          \
 	colon = parser->headerbuf;                                          \
-	while (*colon && *colon != ':')                                   \
-		colon++;                                                  \
-	                                                                  \
+	while (*colon && *colon != ':')                                     \
+		colon++;                                                    \
+	                                                                    \
 	hlen = colon - parser->headerbuf;                                   \
-	                                                                  \
-	header->name = g_strstrip (g_strndup (parser->headerbuf, hlen));    \
-	header->value = g_strstrip (g_strdup (colon + 1));                \
+	                                                                    \
+	header->name = g_strndup (parser->headerbuf, hlen);                 \
+	g_strstrip (header->name);                                          \
+	if (*colon != ':') {                                                \
+		g_warning ("Invalid header: %s", header->name);             \
+		header->value = header->name;                               \
+		header->name = g_strdup ("X-Invalid-Header");               \
+	} else {                                                            \
+		header->value = g_strdup (colon + 1);                       \
+		g_strstrip (header->value);                                 \
+	}                                                                   \
 	header->offset = parser->header_start;                              \
-	                                                                  \
-	hend->next = header;                                              \
-	hend = header;                                                    \
-	                                                                  \
-	parser->headerleft += parser->headerptr - parser->headerbuf;            \
-	parser->headerptr = parser->headerbuf;                                \
+	                                                                    \
+	hend->next = header;                                                \
+	hend = header;                                                      \
+	                                                                    \
+	parser->headerleft += parser->headerptr - parser->headerbuf;        \
+	parser->headerptr = parser->headerbuf;                              \
 } G_STMT_END
 
 static int
@@ -355,7 +366,7 @@ parser_step_headers (GMimeParser *parser)
 	register unsigned char *inptr;
 	unsigned char *start, *inend;
 	struct _header_raw *hend;
-	size_t len;
+	size_t len = 0;
 	
 	parser->midline = FALSE;
 	hend = (struct _header_raw *) &parser->headers;
@@ -363,10 +374,11 @@ parser_step_headers (GMimeParser *parser)
 	parser->header_start = parser_offset (parser, NULL);
 	
 	inptr = parser->inptr;
+	inend = parser->inend;
 	
 	do {
 	refill:
-		if (parser_fill (parser) <= 0)
+		if (parser_fill (parser) <= len)
 			break;
 		
 		inptr = parser->inptr;
@@ -386,6 +398,7 @@ parser_step_headers (GMimeParser *parser)
 				/* we don't have enough data to tell if we
 				   got all of the header or not... */
 				parser->inptr = start;
+				len = inend - start;
 				goto refill;
 			}
 			
@@ -412,6 +425,7 @@ parser_step_headers (GMimeParser *parser)
 		}
 		
 		parser->inptr = inptr;
+		len = inend - inptr;
 	} while (1);
 	
 	inptr = parser->inptr;
@@ -617,16 +631,35 @@ static void
 parser_scan_mime_part_content (GMimeParser *parser, GMimePart *mime_part, int *found)
 {
 	GMimePartEncodingType encoding;
+	GByteArray *content = NULL;
 	GMimeDataWrapper *wrapper;
 	GMimeStream *stream;
 	off_t start, end;
 	
-	start = parser_offset (parser, NULL);
-	*found = parser_scan_content (parser, NULL);
-	end = parser_offset (parser, NULL) - 1;  /* last '\n' belongs to the boundary */
+	if (parser->seekable)
+		start = parser_offset (parser, NULL);
+	else
+		content = g_byte_array_new ();
+	
+	*found = parser_scan_content (parser, content);
+	if (*found != FOUND_EOS) {
+		/* last '\n' belongs to the boundary */
+		if (parser->seekable)
+			end = parser_offset (parser, NULL) - 1;
+		else
+			g_byte_array_set_size (content, MAX (content->len - 1, 0));
+	} else if (parser->seekable) {
+		end = parser_offset (parser, NULL);
+	}
 	
 	encoding = g_mime_part_get_encoding (mime_part);
-	stream = g_mime_stream_substream (parser->stream, start, end);
+	
+	if (parser->seekable) {
+		stream = g_mime_stream_substream (parser->stream, start, end);
+	} else {
+		stream = g_mime_stream_mem_new_with_byte_array (content);
+	}
+	
 	wrapper = g_mime_data_wrapper_new_with_stream (stream, encoding);
 	g_mime_part_set_content_object (mime_part, wrapper);
 	g_mime_stream_unref (stream);
