@@ -30,6 +30,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <errno.h>
 
 #ifdef HAVE_ALLOCA_H
 #include <alloca.h>
@@ -40,8 +41,11 @@
 #include "gmime-part.h"
 #include "gmime-charset.h"
 #include "gmime-iconv.h"
+#include "gmime-iconv-utils.h"
+#include "unicode.h"
 
 #define d(x)
+#define w(x) x
 
 #ifndef HAVE_ISBLANK
 #define isblank(c) (c == ' ' || c == '\t')
@@ -50,10 +54,12 @@
 #define GMIME_UUENCODE_CHAR(c) ((c) ? (c) + ' ' : '`')
 #define	GMIME_UUDECODE_CHAR(c) (((c) - ' ') & 077)
 
+#define GMIME_FOLD_PREENCODED  (GMIME_FOLD_LEN / 2)
+
 /* date parser macros */
 #define NUMERIC_CHARS          "1234567890"
-#define WEEKDAY_CHARS          "Sun,Mon,Tue,Wed,Thu,Fri,Sat,"
-#define MONTH_CHARS            "JanFebMarAprMayJunJulAugSepOctNovDec"
+#define WEEKDAY_CHARS          "SundayMondayTuesdayWednesdayThursdayFridaySaturday"
+#define MONTH_CHARS            "JanuaryFebruaryMarchAprilMayJuneJulyAugustSeptemberOctoberNovemberDecember"
 #define TIMEZONE_ALPHA_CHARS   "UTCGMTESTEDTCSTCDTMSTPSTPDTZAMNY()"
 #define TIMEZONE_NUMERIC_CHARS "-+1234567890"
 #define TIME_CHARS             "1234567890:"
@@ -67,6 +73,9 @@
 #define DATE_TOKEN_NON_TIMEZONE_NUMERIC (1 << 6)
 #define DATE_TOKEN_HAS_SIGN             (1 << 7)
 
+/* from gmime.c */
+extern int gmime_interfaces_utf8;
+
 static char *base64_alphabet =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
@@ -116,12 +125,12 @@ static unsigned char gmime_uu_rank[256] = {
 static unsigned char gmime_datetok_table[256] = {
 	128,111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,
 	111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,
-	111,111,111,111,111,111,111,111, 79, 79,111,175,109,175,111,111,
+	111,111,111,111,111,111,111,111, 79, 79,111,175,111,175,111,111,
 	 38, 38, 38, 38, 38, 38, 38, 38, 38, 38,119,111,111,111,111,111,
 	111, 75,111, 79, 75, 79,105, 79,111,111,107,111,111, 73, 75,107,
 	 79,111,111, 73, 77, 79,111,109,111, 79, 79,111,111,111,111,111,
-	111,105,107,107,109,105,111,107,109,109,111,111,107,111,105,105,
-	107,111,105,111,105,105,107,111,111,107,111,111,111,111,111,111,
+	111,105,107,107,109,105,111,107,105,105,111,111,107,107,105,105,
+	107,111,105,105,105,105,107,111,111,105,111,111,111,111,111,111,
 	111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,
 	111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,
 	111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,111,
@@ -215,7 +224,7 @@ datetok (const char *date)
 		mask = 0;
 		
 		/* find the end of this token */
-		for (end = start; *end && !isspace ((int) *end) && (end > start ? !strchr ("-/", *end) : TRUE); end++) {
+		for (end = start; *end && !strchr ("-/,\t\r\n ", *end); end++) {
 			mask |= gmime_datetok_table[*end];
 		}
 		
@@ -401,22 +410,31 @@ get_time (const unsigned char *in, unsigned int inlen, int *hour, int *min, int
 static int
 get_tzone (struct _date_token **token)
 {
-	int i;
+	const unsigned char *inptr, *inend;
+	unsigned int inlen;
+	int i, t;
 	
 	for (i = 0; *token && i < 2; *token = (*token)->next, i++) {
-		const unsigned char *inptr = (*token)->start;
-		unsigned int inlen = (*token)->len;
+		inptr = (*token)->start;
+		inlen = (*token)->len;
+		inend = inptr + inlen;
 		
 		if (*inptr == '+' || *inptr == '-') {
 			return decode_int (inptr, inlen);
 		} else {
-			int t;
-			
-			if (*inptr == '(')
+			if (*inptr == '(') {
 				inptr++;
+				if (*(inend - 1) == ')')
+					inlen -= 2;
+				else
+					inlen--;
+			}
 			
 			for (t = 0; t < 15; t++) {
-				unsigned int len = MIN (strlen (tz_offsets[t].name), inlen - 1);
+				unsigned int len = strlen (tz_offsets[t].name);
+				
+				if (len != inlen)
+					continue;
 				
 				if (!strncmp (inptr, tz_offsets[t].name, len))
 					return tz_offsets[t].offset;
@@ -486,9 +504,9 @@ parse_rfc822_date (struct _date_token *tokens, int *tzone)
 	}
 	
 	t = mktime (&tm);
-#if defined(HAVE_TIMEZONE)
+#if defined (HAVE_TIMEZONE)
 	t -= timezone;
-#elif defined(HAVE_TM_GMTOFF)
+#elif defined (HAVE_TM_GMTOFF)
 	t += tm.tm_gmtoff;
 #else
 #error Neither HAVE_TIMEZONE nor HAVE_TM_GMTOFF defined. Rerun autoheader, autoconf, etc.
@@ -522,7 +540,7 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 	int hour, min, sec, offset, n;
 	struct _date_token *token;
 	struct tm tm;
-	time_t time;
+	time_t t;
 	
 	memset ((void *) &tm, 0, sizeof (struct tm));
 	got_wday = got_month = got_tzone = FALSE;
@@ -532,7 +550,7 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 	while (token) {
 		if (is_weekday (token) && !got_wday) {
 			if ((n = get_wday (token->start, token->len)) != -1) {
-				printf ("weekday; ");
+				d(printf ("weekday; "));
 				got_wday = TRUE;
 				tm.tm_wday = n;
 				goto next_token;
@@ -541,7 +559,7 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 		
 		if (is_month (token) && !got_month) {
 			if ((n = get_month (token->start, token->len)) != -1) {
-				printf ("month; ");
+				d(printf ("month; "));
 				got_month = TRUE;
 				tm.tm_mon = n;
 				goto next_token;
@@ -550,7 +568,7 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 		
 		if (is_time (token) && !tm.tm_hour && !tm.tm_min && !tm.tm_sec) {
 			if (get_time (token->start, token->len, &hour, &min, &sec)) {
-				printf ("time; ");
+				d(printf ("time; "));
 				tm.tm_hour = hour;
 				tm.tm_min = min;
 				tm.tm_sec = sec;
@@ -562,7 +580,7 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 			struct _date_token *t = token;
 			
 			if ((n = get_tzone (&t)) != -1) {
-				printf ("tzone; ");
+				d(printf ("tzone; "));
 				got_tzone = TRUE;
 				offset = n;
 				goto next_token;
@@ -572,23 +590,23 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 		if (is_numeric (token)) {
 			if (token->len == 4 && !tm.tm_year) {
 				if ((n = get_year (token->start, token->len)) != -1) {
-					printf ("year; ");
+					d(printf ("year; "));
 					tm.tm_year = n - 1900;
 					goto next_token;
 				}
 			} else {
 				if (!got_month && !got_wday && token->next && is_numeric (token->next)) {
-					printf ("mon; ");
+					d(printf ("mon; "));
 					n = decode_int (token->start, token->len);
 					got_month = TRUE;
 					tm.tm_mon = n - 1;
 					goto next_token;
 				} else if (!tm.tm_mday && (n = get_mday (token->start, token->len)) != -1) {
-					printf ("mday; ");
+					d(printf ("mday; "));
 					tm.tm_mday = n;
 					goto next_token;
 				} else if (!tm.tm_year) {
-					printf ("2-digit year; ");
+					d(printf ("2-digit year; "));
 					n = get_year (token->start, token->len);
 					tm.tm_year = n - 1900;
 					goto next_token;
@@ -596,20 +614,20 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 			}
 		}
 		
-		printf ("???; ");
+		d(printf ("???; "));
 		
 	next_token:
 		
 		token = token->next;
 	}
 	
-	printf ("\n");
-		
-	time = mktime (&tm);
-#if defined(HAVE_TIMEZONE)
-	time -= timezone;
-#elif defined(HAVE_TM_GMTOFF)
-	time += tm.tm_gmtoff;
+	d(printf ("\n"));
+	
+	t = mktime (&tm);
+#if defined (HAVE_TIMEZONE)
+	t -= timezone;
+#elif defined (HAVE_TM_GMTOFF)
+	t += tm.tm_gmtoff;
 #else
 #error Neither HAVE_TIMEZONE nor HAVE_TM_GMTOFF defined. Rerun autoheader, autoconf, etc.
 #endif
@@ -617,12 +635,12 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 	/* t is now GMT of the time we want, but not offset by the timezone ... */
 	
 	/* this should convert the time to the GMT equiv time */
-	time -= ((offset / 100) * 60 * 60) + (offset % 100) * 60;
+	t -= ((offset / 100) * 60 * 60) + (offset % 100) * 60;
 	
 	if (tzone)
 		*tzone = offset;
 	
-	return time;
+	return t;
 }
 
 #if 0
@@ -705,6 +723,18 @@ g_mime_utils_header_decode_date (const char *in, int *saveoffset)
 }
 
 
+static void
+g_string_append_len (GString *out, const char *in, size_t len)
+{
+	char *buf;
+	
+	buf = alloca (len + 1);
+	strlcpy (buf, in, len + 1);
+	
+	g_string_append (out, buf);
+}
+
+
 /**
  * g_mime_utils_header_fold:
  * @in: input header string
@@ -716,6 +746,7 @@ g_mime_utils_header_decode_date (const char *in, int *saveoffset)
 char *
 g_mime_utils_header_fold (const char *in)
 {
+	gboolean last_was_lwsp = FALSE;
 	register const char *inptr;
 	size_t len, outlen, i;
 	GString *out;
@@ -731,7 +762,10 @@ g_mime_utils_header_fold (const char *in)
 	while (*inptr) {
 		len = strcspn (inptr, " \t");
 		
-		if (outlen + len > GMIME_FOLD_LEN) {
+		if (outlen + len > GMIME_FOLD_LEN) {			
+			if (last_was_lwsp)
+				g_string_truncate (out, out->len - 1);
+			
 			g_string_append (out, "\n\t");
 			outlen = 1;
 			
@@ -744,21 +778,33 @@ g_mime_utils_header_fold (const char *in)
 				g_string_append (out, "\n\t");
 				outlen = 1;
 			}
+			last_was_lwsp = FALSE;
 		} else if (len > 0) {
 			outlen += len;
-			for (i = 0; i < len; i++) {
-				g_string_append_c (out, inptr[i]);
-			}
+			g_string_append_len (out, inptr, len);
 			inptr += len;
+			last_was_lwsp = FALSE;
 		} else {
-			g_string_append_c (out, *inptr++);
+			if (*inptr == '\t') {
+				/* tabs are a good place to fold, odds
+                                   are that this is where the previous
+                                   mailer folded it */
+				g_string_append (out, "\n\t");
+				outlen = 1;
+				inptr++;
+				last_was_lwsp = FALSE;
+			} else {
+				g_string_append_c (out, *inptr++);
+				outlen++;
+				last_was_lwsp = TRUE;
+			}
 		}
 	}
 	
 	ret = out->str;
 	g_string_free (out, FALSE);
 	
-	return ret;	
+	return ret;
 }
 
 
@@ -956,7 +1002,6 @@ quoted_decode (const unsigned char *in, size_t len, unsigned char *out)
 	register unsigned char *outptr;
 	const unsigned char *inend;
 	unsigned char c, c1;
-	gboolean err = FALSE;
 	
 	inend = in + len;
 	outptr = out;
@@ -972,41 +1017,29 @@ quoted_decode (const unsigned char *in, size_t len, unsigned char *out)
 					| ((c1 >= 'A' ? c1 - 'A' + 10 : c1 - '0') & 0x0f);
 			} else {
 				/* data was truncated */
-				err = TRUE;
-				break;
+				return -1;
 			}
 		} else if (c == '_') {
 			/* _'s are an rfc2047 shortcut for encoding spaces */
 			*outptr++ = ' ';
-		} else if (isblank (c) || strchr (CHARS_ESPECIAL, c)) {
-			/* FIXME: this is an error! ignore for now ... */
-			err = TRUE;
-			break;
 		} else {
 			*outptr++ = c;
 		}
 	}
 	
-	if (!err) {
-		return (outptr - out);
-	}
-	
-	return -1;
+	return (outptr - out);
 }
 
-#define is_8bit_word_encoded(atom, len) (len >= 7 && !strncmp (atom, "=?", 2) && !strncmp (atom + len - 2, "?=", 2))
+#define is_rfc2047_encoded_word(atom, len) (len >= 7 && !strncmp (atom, "=?", 2) && !strncmp (atom + len - 2, "?=", 2))
 
 static unsigned char *
-decode_encoded_8bit_word (const unsigned char *word)
+rfc2047_decode_word (const unsigned char *in, size_t inlen)
 {
 	const register unsigned char *inptr;
 	const unsigned char *inend;
-	size_t len;
 	
-	len = strlen (word);
-	
-	inptr = word + 2;
-	inend = word + len - 2;
+	inptr = in + 2;
+	inend = in + inlen - 2;
 	
 	inptr = memchr (inptr, '?', inend - inptr);
 	if (inptr && inptr[2] == '?') {
@@ -1015,16 +1048,14 @@ decode_encoded_8bit_word (const unsigned char *word)
 		int state = 0;
 		int save = 0;
 		
-		d(fprintf (stderr, "encoding is '%c'\n", inptr[0]));
-		
 		inptr++;
+		
 		switch (*inptr) {
 		case 'B':
 		case 'b':
 			inptr += 2;
 			decoded = alloca (inend - inptr);
 			declen = g_mime_utils_base64_decode_step (inptr, inend - inptr, decoded, &state, &save);
-			return g_strndup (decoded, declen);
 			break;
 		case 'Q':
 		case 'q':
@@ -1036,13 +1067,63 @@ decode_encoded_8bit_word (const unsigned char *word)
 				d(fprintf (stderr, "encountered broken 'Q' encoding\n"));
 				return NULL;
 			}
-			
-			return g_strndup (decoded, declen);
 			break;
 		default:
 			d(fprintf (stderr, "unknown encoding\n"));
 			return NULL;
 		}
+		
+		if (gmime_interfaces_utf8) {
+			const char *charset;
+			unsigned char *buf;
+			char *charenc, *p;
+			size_t len;
+			iconv_t cd;
+			
+			len = (inptr - 3) - (in + 2);
+			charenc = alloca (len + 1);
+			memcpy (charenc, in + 2, len);
+			charenc[len] = '\0';
+			charset = charenc;
+			
+			/* rfc2231 updates rfc2047 encoded words...
+			 * The ABNF given in RFC 2047 for encoded-words is:
+			 *   encoded-word := "=?" charset "?" encoding "?" encoded-text "?="
+			 * This specification changes this ABNF to:
+			 *   encoded-word := "=?" charset ["*" language] "?" encoding "?" encoded-text "?="
+			 */
+			
+			/* trim off the 'language' part if it's there... */
+			p = strchr (charset, '*');
+			if (p)
+				*p = '\0';
+			
+			/* slight optimization */
+			if (!strcasecmp (charset, "UTF-8"))
+				return g_strndup (decoded, declen);
+			
+			cd = g_mime_iconv_open ("UTF-8", charset);
+			if (cd == (iconv_t) -1) {
+				w(g_warning ("Cannot convert from %s to UTF-8, header display may "
+					     "be corrupt: %s", charset, g_strerror (errno)));
+				charset = g_mime_charset_locale_name ();
+				cd = g_mime_iconv_open ("UTF-8", charset);
+				if (cd == (iconv_t) -1)
+					return NULL;
+			}
+			
+			buf = g_mime_iconv_strndup (cd, decoded, declen);
+			g_mime_iconv_close (cd);
+			
+			if (!buf) {
+				w(g_warning ("Failed to convert \"%.*s\" to UTF-8, display may be "
+					     "corrupt: %s", declen, decoded, g_strerror (errno)));
+			}
+			
+			return buf;
+		} else {
+			return g_strndup (decoded, declen);
+		}
 	}
 	
 	return NULL;
@@ -1080,8 +1161,8 @@ g_mime_utils_8bit_header_decode (const unsigned char *in)
 			const unsigned char *word;
 			gboolean was_encoded;
 			
-			if ((was_encoded = is_8bit_word_encoded (atom->str, atom->len)))
-				word = dword = decode_encoded_8bit_word (atom->str);
+			if ((was_encoded = is_rfc2047_encoded_word (atom->str, atom->len)))
+				word = dword = rfc2047_decode_word (atom->str, atom->len);
 			else
 				word = atom->str;
 			
@@ -1135,8 +1216,8 @@ g_mime_utils_8bit_header_decode (const unsigned char *in)
 		const unsigned char *word;
 		gboolean was_encoded;
 		
-		if ((was_encoded = is_8bit_word_encoded (atom->str, atom->len)))
-			word = dword = decode_encoded_8bit_word (atom->str);
+		if ((was_encoded = is_rfc2047_encoded_word (atom->str, atom->len)))
+			word = dword = rfc2047_decode_word (atom->str, atom->len);
 		else
 			word = atom->str;
 		
@@ -1194,16 +1275,34 @@ quoted_encode (const unsigned char *in, size_t len, unsigned char *out, gushort
 	return (outptr - out);
 }
 
-static char *
-encode_8bit_word (const unsigned char *word, gushort safemask, gboolean *this_was_encoded)
+static void
+rfc2047_encode_word (GString *string, const unsigned char *word, size_t len,
+		     const char *charset, gushort safemask)
 {
 	unsigned char *encoded, *ptr;
-	size_t enclen, pos, len;
+	unsigned char *uword = NULL;
+	iconv_t cd = (iconv_t) -1;
+	size_t enclen, pos;
 	int state = 0;
 	int save = 0;
 	char encoding;
 	
-	len = strlen (word);
+	if (gmime_interfaces_utf8) {
+		if (strcasecmp (charset, "UTF-8") != 0)
+			cd = g_mime_iconv_open (charset, "UTF-8");
+		
+		if (cd != (iconv_t) -1) {
+			uword = g_mime_iconv_strndup (cd, word, len);
+			g_mime_iconv_close (cd);
+		}
+		
+		if (uword) {
+			len = strlen (uword);
+			word = uword;
+		} else {
+			charset = "UTF-8";
+		}
+	}
 	
 	switch (g_mime_utils_best_encoding (word, len)) {
 	case GMIME_PART_ENCODING_BASE64:
@@ -1232,16 +1331,12 @@ encode_8bit_word (const unsigned char *word, gushort safemask, gboolean *this_wa
 		
 		break;
 	default:
-		if (this_was_encoded)
-			*this_was_encoded = FALSE;
-		
-		return g_strdup (word);
+		g_assert_not_reached ();
 	}
 	
-	if (this_was_encoded)
-		*this_was_encoded = TRUE;
+	g_free (uword);
 	
-	return g_strdup_printf ("=?%s?%c?%s?=", g_mime_charset_locale_name (), encoding, encoded);	
+	g_string_sprintfa (string, "=?%s?%c?%s?=", charset, encoding, encoded);
 }
 
 
@@ -1257,7 +1352,254 @@ encode_8bit_word (const unsigned char *word, gushort safemask, gboolean *this_wa
 char *
 g_mime_utils_8bit_header_encode_phrase (const unsigned char *in)
 {
-	return encode_8bit_word (in, IS_PSAFE, NULL);
+	const char *charset;
+	GString *string;
+	size_t len;
+	char *str;
+	
+	if (in == NULL)
+		return NULL;
+	
+	len = strlen (in);
+	
+	if (gmime_interfaces_utf8) {
+		charset = g_mime_charset_best (in, len);
+		charset = charset ? charset : "iso-8859-1";
+	} else {
+		charset = g_mime_charset_locale_name ();
+	}
+	
+	string = g_string_new ("");
+	
+	rfc2047_encode_word (string, in, strlen (in), charset, IS_ESAFE);
+	
+	str = string->str;
+	g_string_free (string, FALSE);
+	
+	return str;
+}
+
+
+enum _phrase_word_t {
+	WORD_ATOM,
+	WORD_2047
+};
+
+struct _phrase_word {
+	struct _phrase_word *next;
+	const unsigned char *start, *end;
+	enum _phrase_word_t type;
+	int encoding;
+};
+
+static gboolean
+word_types_compatable (enum _phrase_word_t type1, enum _phrase_word_t type2)
+{
+	switch (type1) {
+	case WORD_ATOM:
+		return FALSE;
+	case WORD_2047:
+		return type2 == WORD_2047;
+	default:
+		return FALSE;
+	}
+}
+
+static struct _phrase_word *
+rfc2047_encode_phrase_get_words (const unsigned char *in)
+{
+	const unsigned char *inptr, *start, *last;
+	struct _phrase_word *words, *tail, *word;
+	enum _phrase_word_t type = WORD_ATOM;
+	int count = 0, encoding = 0;
+	
+	words = NULL;
+	tail = (struct _phrase_word *) &words;
+	
+	last = start = inptr = in;
+	while (inptr && *inptr) {
+		gboolean is_space;
+		unichar c;
+		
+		if (gmime_interfaces_utf8) {
+			const char *newinptr;
+			
+			newinptr = unicode_next_char (inptr);
+			c = unicode_get_char (inptr);
+			if (newinptr == NULL || !unichar_validate (c)) {
+				w(g_warning ("Invalid UTF-8 sequence encountered"));
+				inptr++;
+				continue;
+			}
+			
+			inptr = newinptr;
+			
+			is_space = unichar_isspace (c);
+		} else {
+			is_space = isspace ((int) *inptr);
+			c = *inptr++;
+		}
+		
+		if (is_space) {
+			if (count > 0) {
+				word = g_new (struct _phrase_word, 1);
+				word->next = NULL;
+				word->start = start;
+				word->end = last;
+				word->type = type;
+				word->encoding = encoding;
+				
+				tail->next = word;
+				tail = word;
+				count = 0;
+			}
+			
+			start = inptr;
+			type = WORD_ATOM;
+			encoding = 0;
+		} else {
+			count++;
+			if (c > 127 && c < 256) {
+				type = WORD_2047;
+				encoding = MAX (encoding, 2);
+			} else if (c >= 256) {
+				type = WORD_2047;
+				encoding = 2;
+			}
+		}
+		
+		last = inptr;
+	}
+	
+	if (count > 0) {
+		word = g_new (struct _phrase_word, 1);
+		word->next = NULL;
+		word->start = start;
+		word->end = last;
+		word->type = type;
+		word->encoding = encoding;
+		
+		tail->next = word;
+		tail = word;
+	}
+	
+	return words;
+}
+
+static gboolean
+rfc2047_encode_phrase_merge_words (struct _phrase_word **wordsp)
+{
+	struct _phrase_word *word, *next, *words = *wordsp;
+	gboolean merged = FALSE;
+	
+	/* scan the list, checking for words of similar types that can be merged */
+	word = words;
+	while (word) {
+		next = word->next;
+		
+		while (next) {
+			/* merge nodes of the same type AND we are not creating too long a string */
+			if (word_types_compatable (word->type, next->type)) {
+				if (next->end - word->start < GMIME_FOLD_PREENCODED) {
+					/* the resulting word type is the MAX of the 2 types */
+					word->type = MAX (word->type, next->type);
+					
+					word->end = next->end;
+					word->next = next->next;
+					
+					g_free (next);
+					
+					next = word->next;
+					
+					merged = TRUE;
+				} else {
+					/* if it is going to be too long, make sure we include the
+					   separating whitespace */
+					word->end = next->start;
+					break;
+				}
+			} else {
+				break;
+			}
+		}
+		
+		word = word->next;
+	}
+	
+	*wordsp = words;
+	
+	return merged;
+}
+
+static char *
+rfc2047_encode_phrase (const unsigned char *in)
+{
+	struct _phrase_word *words, *word, *prev = NULL;
+	GString *out;
+	char *outstr;
+	
+	if (in == NULL)
+		return NULL;
+	
+	words = rfc2047_encode_phrase_get_words (in);
+	if (!words)
+		return NULL;
+	
+	while (rfc2047_encode_phrase_merge_words (&words))
+		;
+	
+	out = g_string_new ("");
+	
+	/* output words now with spaces between them */
+	word = words;
+	while (word) {
+		const char *start;
+		size_t len;
+		
+		/* append correct number of spaces between words */
+		if (prev && !(prev->type == WORD_2047 && word->type == WORD_2047)) {
+			/* one or both of the words are not encoded so we write the spaces out untouched */
+			len = word->start - prev->end;
+			g_string_append_len (out, prev->end, len);
+		}
+		
+		switch (word->type) {
+		case WORD_ATOM:
+			g_string_append_len (out, word->start, word->end - word->start);
+			break;
+		case WORD_2047:
+			if (prev && prev->type == WORD_2047) {
+				/* include the whitespace chars between these 2 words in the
+                                   resulting rfc2047 encoded word. */
+				len = word->end - prev->end;
+				start = prev->end;
+				
+				/* encoded words need to be separated by linear whitespace */
+				g_string_append_c (out, ' ');
+			} else {
+				len = word->end - word->start;
+				start = word->start;
+			}
+			
+			if (word->encoding == 1 || !gmime_interfaces_utf8)
+				rfc2047_encode_word (out, start, len, "iso-8859-1", IS_PSAFE);
+			else
+				rfc2047_encode_word (out, start, len,
+						     g_mime_charset_best (start, len), IS_PSAFE);
+			break;
+		}
+		
+		g_free (prev);
+		prev = word;
+		word = word->next;
+	}
+	
+	g_free (prev);
+	
+	outstr = out->str;
+	g_string_free (out, FALSE);
+	
+	return outstr;
 }
 
 
@@ -1273,104 +1615,7 @@ g_mime_utils_8bit_header_encode_phrase (const unsigned char *in)
 char *
 g_mime_utils_8bit_header_encode (const unsigned char *in)
 {
-	register const unsigned char *inptr;
-	GString *out, *word, *lwsp;
-	unsigned char *encoded;
-	gboolean is8bit = FALSE;
-	gboolean last_was_encoded = FALSE;
-	gboolean last_was_space = FALSE;
-	
-	out = g_string_new ("");
-	word = g_string_new ("");
-	lwsp = g_string_new ("");
-	inptr = (guchar *) in;
-	
-	while (inptr && *inptr) {
-		unsigned char c = *inptr++;
-		
-		if (isspace (c) && !last_was_space) {
-			gboolean this_was_encoded = FALSE;
-			char *eword;
-			
-			if (is8bit)
-				eword = encode_8bit_word (word->str, IS_ESAFE, &this_was_encoded);
-			else
-				eword = g_strdup (word->str);
-			
-			/* append any whitespace */
-			if (last_was_encoded && this_was_encoded) {
-				/* we need to encode the whitespace */
-				unsigned char *elwsp;
-				size_t len;
-				
-				elwsp = alloca (lwsp->len * 3 + 4);
-				len = quoted_encode (lwsp->str, lwsp->len, elwsp, IS_SPACE);
-				elwsp[len] = '\0';
-				
-				g_string_sprintfa (out, " =?%s?q?%s?= ", g_mime_charset_locale_name (), elwsp);
-			} else {
-				g_string_append (out, lwsp->str);
-			}
-			
-			/* append the encoded word */
-			g_string_append (out, eword);
-			g_free (eword);
-			
-			g_string_truncate (lwsp, 0);
-			g_string_truncate (word, 0);
-			
-			last_was_encoded = this_was_encoded;
-			is8bit = FALSE;
-		}
-		
-		if (isspace (c)) {
-			g_string_append_c (lwsp, c);
-			last_was_space = TRUE;
-		} else {
-			if (c > 127)
-				is8bit = TRUE;
-			
-			g_string_append_c (word, c);
-			last_was_space = FALSE;
-		}
-	}
-	
-	if (word->len || lwsp->len) {
-		gboolean this_was_encoded = FALSE;
-		char *eword;
-		
-		if (is8bit)
-			eword = encode_8bit_word (word->str, IS_ESAFE, &this_was_encoded);
-		else
-			eword = g_strdup (word->str);
-		
-		/* append any whitespace */
-		if (last_was_encoded && this_was_encoded) {
-			/* we need to encode the whitespace */
-			unsigned char *elwsp;
-			size_t len;
-			
-			elwsp = alloca (lwsp->len * 3 + 4);
-			len = quoted_encode (lwsp->str, lwsp->len, elwsp, IS_SPACE);
-			elwsp[len] = '\0';
-			
-			g_string_sprintfa (out, " =?%s?q?%s?= ", g_mime_charset_locale_name (), elwsp);
-		} else {
-			g_string_append (out, lwsp->str);
-		}
-		
-		/* append the encoded word */
-		g_string_append (out, eword);
-		g_free (eword);
-	}
-	
-	g_string_free (lwsp, TRUE);
-	g_string_free (word, TRUE);
-	
-	encoded = out->str;
-	g_string_free (out, FALSE);
-	
-	return encoded;
+	return rfc2047_encode_phrase (in);
 }
 
 
@@ -1450,8 +1695,6 @@ g_mime_utils_base64_encode_step (const unsigned char *in, size_t inlen, unsigned
 	inptr = in;
 	outptr = out;
 	
-	d(printf("we have %d chars, and %d saved chars\n", inlen, ((gchar *)save)[0]));
-	
 	if (inlen + ((unsigned char *)save)[0] > 2) {
 		const unsigned char *inend = in + inlen - 2;
 		register int c1 = 0, c2 = 0, c3 = 0;
@@ -1564,7 +1807,7 @@ g_mime_utils_base64_decode_step (const unsigned char *in, size_t inlen, unsigned
 	while (inptr > in && i) {
 		inptr--;
 		if (gmime_base64_rank[*inptr] != 0xff) {
-			if (*inptr == '=')
+			if (*inptr == '=' && outptr > out)
 				outptr--;
 			i--;
 		}
@@ -1594,13 +1837,15 @@ g_mime_utils_uuencode_close (const unsigned char *in, size_t inlen, unsigned cha
 {
 	register unsigned char *outptr, *bufptr;
 	register guint32 saved;
-	int uulen, i;
+	int uulen, uufill, i;
 	
 	outptr = out;
 	
 	if (inlen > 0)
 		outptr += g_mime_utils_uuencode_step (in, inlen, out, uubuf, state, save);
 	
+	uufill = 0;
+	
 	saved = *save;
 	i = *state & 0xff;
 	uulen = (*state >> 8) & 0xff;
@@ -1610,6 +1855,7 @@ g_mime_utils_uuencode_close (const unsigned char *in, size_t inlen, unsigned cha
 	if (i > 0) {
 		while (i < 3) {
 			saved <<= 8 | 0;
+			uufill++;
 			i++;
 		}
 		
@@ -1635,7 +1881,7 @@ g_mime_utils_uuencode_close (const unsigned char *in, size_t inlen, unsigned cha
 	if (uulen > 0) {
 		int cplen = ((uulen / 3) * 4);
 		
-		*outptr++ = GMIME_UUENCODE_CHAR (uulen & 0xff);
+		*outptr++ = GMIME_UUENCODE_CHAR ((uulen - uufill) & 0xff);
 		memcpy (outptr, uubuf, cplen);
 		outptr += cplen;
 		*outptr++ = '\n';
@@ -1771,17 +2017,19 @@ g_mime_utils_uudecode_step (const unsigned char *in, size_t inlen, unsigned char
 	outptr = out;
 	
 	inptr = in;
-	while (inptr < inend && *inptr) {
-		if (*inptr == '\n' || last_was_eoln) {
-			if (last_was_eoln) {
-				uulen = gmime_uu_rank[*inptr];
-				last_was_eoln = FALSE;
-				if (uulen == 0) {
-					*state |= GMIME_UUDECODE_STATE_END;
-					break;
-				}
-			} else {
-				last_was_eoln = TRUE;
+	while (inptr < inend) {
+		if (*inptr == '\n') {
+			last_was_eoln = TRUE;
+			
+			inptr++;
+			continue;
+		} else if (!uulen || last_was_eoln) {
+			/* first octet on a line is the uulen octet */
+			uulen = gmime_uu_rank[*inptr];
+			last_was_eoln = FALSE;
+			if (uulen == 0) {
+				*state |= GMIME_UUDECODE_STATE_END;
+				break;
 			}
 			
 			inptr++;
